[<Microsoft.FSharp.Core.AutoOpen>]
module Degenz.Types

open System

[<Measure>]
type mins

[<Measure>]
type GBT

type Range = { Min : int ; Max : int }

module Range =
    let normalized = { Min = 0 ; Max = 100 }
    let create min max = { Min = min ; Max = max }
    let constrain value range = if value < range.Min then range.Min elif value > range.Max then range.Max else value

type ItemId =
    | VIRUS = 0
    | REMOTE = 1
    | WORM = 2
    | FIREWALL = 6
    | ENCRYPTION = 7
    | CYPHER = 8

type StatId =
    | Strength = 0
    | Focus = 1
    | Charisma = 2
    | Luck = 3

type StatConfig = {
    Id : StatId
    BaseDecayRate : float
    BaseRange : Range
}

type PlayerStat = {
    Id : StatId
    Amount : int
    ModRange : Range
    LastRead : DateTime
}
with static member empty = { Id = StatId.Strength ; Amount = 0 ; ModRange = Range.normalized ; LastRead = DateTime.UtcNow}

type Stats = {
    Strength : PlayerStat
    Focus : PlayerStat
    Luck : PlayerStat
    Charisma : PlayerStat
}
with static member empty = { Strength = PlayerStat.empty ; Focus = PlayerStat.empty ; Luck = PlayerStat.empty ; Charisma = PlayerStat.empty }

type HackResult =
    | Strong
    | Weak

type DiscordPlayer = { Id: uint64; Name: string }
    with static member empty = { Id = 0uL ; Name = "None" }

type HackEvent = {
    IsInstigator : bool
    Adversary : DiscordPlayer
    Success : bool
    HackId : string
}

type PlayerEventType =
    | Hacking of HackEvent
    | Shielding of shieldId : string
    | Stealing of instigator : bool * adversary : DiscordPlayer
    | PlayingSlot
    | Imprison

type PlayerEvent =
    { Type : PlayerEventType
      Cooldown : int<mins>
      Timestamp : DateTime }

[<RequireQualifiedAccess>]
type ItemType =
    | Hack
    | Shield
    | Food
    | Accessory
    | Jpeg
    | Whitelist
    | Misc
    
type Effect =
    | Min of int
    | Max of int
    | Add of int
    | RateMultiplier of float
    
type StatEffect = {
    TargetStat : StatId
    Effect : Effect
}

type ItemAttribute =
    | Buyable of price : int<GBT>
    | Sellable of price : int<GBT>
    | RateLimitable of cooldown : int<mins>
    | Expireable of lifetime : int<mins>
    | Consumable
    | Modifiable of effects : StatEffect list
    | Droppable of chance : float
    | Tradeable
    | Attackable of power : int
    | Defendable of resistance : int
    | Classable of className : string
    | Stackable of max : int

let (|CanBuy|_|) itemAttrs       = itemAttrs |> List.tryPick (function Buyable       p  -> Some p  | _ -> None)
let (|CanSell|_|) itemAttrs      = itemAttrs |> List.tryPick (function Sellable      p  -> Some p  | _ -> None)
let (|CanExpire|_|) itemAttrs    = itemAttrs |> List.tryPick (function Expireable    l  -> Some l  | _ -> None)
let (|CanRateLimit|_|) itemAttrs = itemAttrs |> List.tryPick (function RateLimitable l  -> Some l  | _ -> None)
let (|CanConsume|_|) itemAttrs   = itemAttrs |> List.tryPick (function Consumable       -> Some () | _ -> None)
let (|CanModify|_|) itemAttrs    = itemAttrs |> List.tryPick (function Modifiable    es -> Some es | _ -> None)
let (|CanDrop|_|) itemAttrs      = itemAttrs |> List.tryPick (function Droppable     c  -> Some c  | _ -> None)
let (|CanTrade|_|) itemAttrs     = itemAttrs |> List.tryPick (function Tradeable        -> Some () | _ -> None)
let (|CanAttack|_|) itemAttrs    = itemAttrs |> List.tryPick (function Attackable    p  -> Some p  | _ -> None)
let (|CanDefend|_|) itemAttrs    = itemAttrs |> List.tryPick (function Defendable    r  -> Some r  | _ -> None)
let (|CanClass|_|) itemAttrs     = itemAttrs |> List.tryPick (function Classable     c  -> Some c  | _ -> None)
let (|CanStack|_|) itemAttrs     = itemAttrs |> List.tryPick (function Stackable     m  -> Some m  | _ -> None)
    
type Item = {
    Id : string
    Name : string
    Description : string
    Type : ItemType
    IconUrl : string
    ImageUrl : string
    Attributes : ItemAttribute list
}

type StoreItem = {
    StoreId : string
    Stock : int
    LimitStock : bool
    Available : bool
    Item : Item
}

type HackItem = {
    Id : string
    Name : string
    Price : int<GBT>
    Cooldown : int<mins>
    IconUrl : string
    ImageUrl : string
    Power : int
    Class : string
}

type ShieldItem = {
    Id : string
    Name : string
    Price : int<GBT>
    Cooldown : int<mins>
    IconUrl : string
    ImageUrl : string
    Resistance : int
    Class : string
}

type Inventory = Item list

type PlayerData = {
    DiscordId : uint64
    Name : string
    Inventory : Inventory
    Events : PlayerEvent list
    Stats : Stats
    Bank : int<GBT>
    Active : bool
}
//        Achievements : string array
//        XP : int
with member this.toDiscordPlayer = { Id = this.DiscordId ; Name = this.Name }
     static member empty =
         { DiscordId = 0uL
           Name = "None"
           Inventory = []
           Events = []
           Stats = Stats.empty
//               Achievements = [||]
//               XP = 0
           Bank = 0<GBT>
           Active = false }

module PlayerStats =
    // 2.09f would go from 100 to 0 in roughly 48 hours
    let Strength = { Id = StatId.Strength ; BaseDecayRate = 2.09 ; BaseRange = Range.normalized }
    let Focus    = { Id = StatId.Focus ; BaseDecayRate = 2.09 ; BaseRange = Range.normalized }
    let Luck     = { Id = StatId.Luck ; BaseDecayRate = 2.09 ; BaseRange = Range.normalized }
    let Charisma = { Id = StatId.Charisma ; BaseDecayRate = 2.09 ; BaseRange = Range.normalized }

    let stats = [ Strength ; Focus ; Charisma ; Luck ]

    let getPlayerStat (statConfig : StatConfig) player =
        match statConfig.Id with
        | StatId.Strength -> player.Stats.Strength
        | StatId.Focus -> player.Stats.Focus
        | StatId.Charisma -> player.Stats.Charisma
        | StatId.Luck -> player.Stats.Luck
        | _ -> player.Stats.Luck
    
    let calculateActiveStat statId amount items =
        let statConfig = stats |> List.find (fun s -> s.Id = statId)
//        let hoursElapsed = (DateTime.UtcNow - lastRead).Hours
//        let totalDecay = float hoursElapsed * statConfig.BaseDecayRate
        let modMinMax =
            let min =
                items
                |> List.choose (fun i -> match i.Attributes with CanModify fx -> Some fx | _ -> None)
                |> List.concat
                |> List.sumBy (fun fx -> match fx.Effect with | Min x -> x | _ -> 0)
            let max =
                items
                |> List.choose (fun i -> match i.Attributes with CanModify fx -> Some fx | _ -> None)
                |> List.concat
                |> List.sumBy (fun fx -> match fx.Effect with | Max x -> x | _ -> 0)
            Range.create (statConfig.BaseRange.Min + min) (statConfig.BaseRange.Max + max)
        let amountAfterDecay = modMinMax |> Range.constrain amount
        { Id = statId ; Amount = amountAfterDecay ; ModRange = modMinMax ; LastRead = DateTime.UtcNow }
